"""Module for account class."""


class Account:
    """Account class for the Google API."""

    def __init__(self, scopes: list, service_account_key_path: str = None):
        """Initialize the account."""
        self.scopes = scopes
        self.service_account_key_path = service_account_key_path



"""Module for working with Google Sheets."""

from googleapiclient.discovery import build
from googleapiclient.errors import Error as GoogleError
from gspread.utils import MergeType, a1_range_to_grid_range, rowcol_to_a1
from retry import retry

from ..columns_blocks import ColumnsBlock
from ..goole_api.account import Account
from ..goole_api.goolge_services import GoogleServices
from ..models import DumpCell


class GoogleSheet(GoogleServices):
    """Google Sheet class."""

    def __init__(self, account: Account) -> None:
        """Initialize the Google Sheet."""
        self.service = build("sheets", "v4", credentials=self._get_credentials(account), cache_discovery=False)
        self.header = []
        self.sub_header = []
        self.body = {
            "requests": [],
        }
        self.body["requests"].append(
            {
                "updateSheetProperties": {
                    "properties": {"title": "Test Runs", "grid_properties": {"frozen_row_count": 2}},
                    "fields": "title,gridProperties.frozenRowCount",
                }
            }
        )
        self._file_id = None

    def set_file_id(self, file_id: str) -> None:
        """Set the file id."""
        self._file_id = file_id

    def get_file_id(self) -> str:
        """Get the file id."""
        return self._file_id

    @retry(tries=3, delay=1, exceptions=GoogleError)
    def get_remote_header_block(self):
        """Get headers from google sheet."""
        remote_headers = self.service.spreadsheets().values().get(spreadsheetId=self._file_id, range="1:2").execute()
        return remote_headers.get("values", [])

    def _check_or_create_headers(self, headers: list[DumpCell]) -> None:
        """Check or create the headers."""
        remote_header_block = self.get_remote_header_block()
        if not remote_header_block:
            return self.create_header(self._file_id)
        else:
            headers.sort(key=lambda x: x.column_number)
            remote_sub_headers: list = remote_header_block[1]
            for index, header in enumerate(headers, start=1):
                if header.value not in remote_sub_headers:
                    self._create_col(index, header.value)
            self._format_headers_block()

    def _create_col(self, index: int, value) -> None:
        """Create the column."""
        self._insert_row(index=index)
        self._set_header_value(index=index, value=value)

    @retry(tries=3, delay=1, exceptions=GoogleError)
    def _set_header_value(self, index: int, value: str) -> None:
        self.service.spreadsheets().values().update(
            spreadsheetId=self._file_id,
            range=rowcol_to_a1(2, index),
            valueInputOption="RAW",
            body={"values": [[value]]},
        ).execute()

    @retry(tries=3, delay=1, exceptions=GoogleError)
    def _insert_row(self, index: int) -> None:
        self.service.spreadsheets().batchUpdate(
            spreadsheetId=self._file_id,
            body={
                "requests": [
                    {
                        "insertDimension": {
                            "range": {
                                "dimension": "COLUMNS",
                                "startIndex": index - 1,
                                "endIndex": index,
                            },
                            "inheritFromBefore": True,
                        }
                    }
                ]
            },
        ).execute()

    @retry(tries=3, delay=1, exceptions=GoogleError)
    def is_header_exist(self, file_id: str) -> bool:
        """Check the headers."""
        header = self.service.spreadsheets().values().get(spreadsheetId=file_id, range="1:2").execute()
        header = header.get("values", [])
        if not header:
            return False
        for row in header:
            if not row:
                return False
        else:
            return True

    def get_range(
        self,
        start_row: int,
        start_column: int,
        end_row: int,
        end_column: int,
    ) -> dict:
        """Get the range."""
        range_start = rowcol_to_a1(start_row, start_column)
        range_end = rowcol_to_a1(end_row, end_column)
        return a1_range_to_grid_range(f"{range_start}:{range_end}")

    @retry(tries=3, delay=1, exceptions=GoogleError)
    def create_header(self, file_id: str) -> None:
        """Create the header."""
        self.service.spreadsheets().values().update(
            spreadsheetId=file_id,
            range="1:2",
            valueInputOption="RAW",
            body={"values": [self.header, self.sub_header]},
        ).execute()

    def get_header_block(self, block: ColumnsBlock, block_start_column: int = 0) -> None:
        """Get the header block."""
        if not block.column_names:
            return
        block_end_column = block_start_column + len(block.column_names) - 1
        merge_range = self.get_range(1, block_start_column, 1, block_end_column)
        format_range = self.get_range(1, block_start_column, 2, block_end_column)
        self.body["requests"].append({"mergeCells": {"mergeType": MergeType.merge_all, "range": merge_range}})
        alignment = self.get_aligments("CENTER", "MIDDLE")
        bg_color = self.get_bg_color(block.color)
        text_format = self.get_text_format(font_size=10, bold=True, italic=False)
        self.body["requests"].append(
            {
                "repeatCell": {
                    "range": format_range,
                    "cell": {
                        "userEnteredFormat": {
                            **alignment,
                            **bg_color,
                            **text_format,
                        }
                    },
                    "fields": "userEnteredFormat(horizontalAlignment, verticalAlignment, textFormat, backgroundColor)",
                }
            }
        )
        header = [""] * len(block.column_names)
        header[0] = block.name
        self.header.extend(header)
        self.sub_header.extend(block.column_names)

    def write_to_google_sheet(self, all_cells: list, headers: list, row_number: int) -> None:
        """Write to the Google Sheet."""
        values = [None] * len(headers)
        for cell in all_cells:
            values[cell.column_number - 1] = cell.value
        for cell in all_cells:
            format_range = self.get_range(row_number, cell.column_number, row_number, cell.column_number)
            alignment = self.get_aligments(cell.alignment, "MIDDLE")
            bg_color = self.get_bg_color(cell.bg_color)
            self.body["requests"].append(
                {
                    "repeatCell": {
                        "range": format_range,
                        "cell": {"userEnteredFormat": {**alignment, **bg_color, "wrapStrategy": "WRAP"}},
                        "fields": "userEnteredFormat(horizontalAlignment, verticalAlignment,"
                        " backgroundColor, wrapStrategy)",
                    }
                }
            )
        self._add_row_to_sheet(row_number=row_number, values=values)
        self.body["requests"].append(
            {
                "setBasicFilter": {
                    "filter": {
                        "range": {
                            "startRowIndex": 1,
                            "endRowIndex": row_number,
                            "startColumnIndex": 0,
                            "endColumnIndex": len(headers),
                        }
                    }
                }
            }
        )
        self._update_google_sheet_styles()

    @retry(tries=3, delay=1, exceptions=GoogleError)
    def _update_google_sheet_styles(self):
        self.service.spreadsheets().batchUpdate(spreadsheetId=self._file_id, body=self.body).execute()

    @retry(tries=3, delay=1, exceptions=GoogleError)
    def _add_row_to_sheet(self, row_number: int, values: list):
        self.service.spreadsheets().values().update(
            spreadsheetId=self._file_id,
            range=f"{row_number}:{row_number}",
            valueInputOption="USER_ENTERED",
            body={
                "values": [values],
            },
        ).execute()

    def get_last_row(self) -> int:
        """Get the last row."""
        header = self.service.spreadsheets().values().get(spreadsheetId=self._file_id, range="A:A").execute()
        last_row = len(header.get("values", [])) + 1
        return last_row

    def get_aligments(self, vertical: str, horizontal: str) -> dict:
        """Get the alignments block."""
        return {
            "horizontalAlignment": vertical,
            "verticalAlignment": horizontal,
        }

    def get_bg_color(self, color: str) -> dict:
        """Get the background color block."""
        if color is None:
            return {}
        hex_color = color[-6:]
        red, green, blue = int(hex_color[:2], 16), int(hex_color[2:4], 16), int(hex_color[4:], 16)
        red = red / 255
        green = green / 255
        blue = blue / 255
        return {"backgroundColor": {"red": red, "green": green, "blue": blue}}

    def get_text_format(self, font_size: int, bold: bool, italic: bool) -> dict:
        """Get the text format block."""
        return {
            "textFormat": {
                "fontSize": font_size,
                "bold": bold,
                "italic": italic,
            }
        }

    @retry(tries=3, delay=1, exceptions=GoogleError)
    def _format_headers_block(self) -> None:
        remote_header_block = self.get_remote_header_block()
        start_col_indexes = [remote_header_block[0].index(header) for header in remote_header_block[0] if header]
        end_col_indexes = start_col_indexes[1:]
        end_col_indexes.append(len(remote_header_block[1]))
        body = {
            "requests": [],
        }
        for end_col_index, start_col_index in zip(end_col_indexes, start_col_indexes):
            merge_range = {
                "startRowIndex": 0,
                "endRowIndex": 1,
                "startColumnIndex": start_col_index,
                "endColumnIndex": end_col_index,
            }
            body["requests"].append({"mergeCells": {"mergeType": MergeType.merge_all, "range": merge_range}})

        self.service.spreadsheets().batchUpdate(spreadsheetId=self._file_id, body=body).execute()



"""Module for working with Google Drive API."""
import json
import logging
from time import sleep

from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from googleapiclient.http import MediaFileUpload
from retry import retry

from ..goole_api.account import Account
from ..goole_api.goolge_services import GoogleServices
from ..logger import logger


class MimeTypes:
    """Mime Types class."""

    SPREADSHEET = "application/vnd.google-apps.spreadsheet"
    FOLDER = "application/vnd.google-apps.folder"
    JSON = "application/json"


class GoogleDriveService(GoogleServices):
    """Google Drive Service class."""

    def __init__(self, account: Account):
        """Initialize the Google Drive Service."""
        self.service = build("drive", "v3", credentials=self._get_credentials(account), cache_discovery=False)

    def create_folder_if_not_exists(self, folder_name: str) -> dict:
        """Create the folder if not exists."""
        root_folder = self.get_root_folder("QA - Run Reports")
        try:
            folder = self.get_child_folder_by_name(root_folder, folder_name)
            logger.info(f"Folder {folder_name} already exists")
            return folder
        except ValueError:
            folder = self.create_folder(folder_name=folder_name, parent=[root_folder])
            logger.info(f"Folder {folder_name} created")
            return folder

    def update_json_file(self, file: dict, data: dict) -> dict:
        """Update the JSON file."""
        with open("updated_file.json", "w") as temp_file:
            json.dump(data, temp_file)
        media = MediaFileUpload("updated_file.json", mimetype="application/json")
        self.service.files().update(fileId=file["id"], media_body=media).execute()
        return file

    def create_file_if_not_exists(self, folder: dict, file_name: str, mime_type=MimeTypes.SPREADSHEET) -> dict:
        """Create the file if not exists."""
        try:
            file = self.get_file_in_folder(folder, file_name, mime_type)
            logger.info(f"File {file['name']} already exists")
            return file
        except ValueError:
            file = self.create_file(file_name=file_name, folder=folder, mime_type=mime_type)
            logger.info(f"File {file['name']} created")
            return file

    def get_file_in_folder(self, parents: dict, file_name: str, mime_type: str) -> dict:
        """Get the Google Sheet in the folder."""
        query = f"'{parents['id']}' in parents and mimeType='{mime_type}'"
        files = self.__execute_get_query(query, ["id", "name"])
        for file in files:
            if file["name"] == file_name:
                return file
        else:
            raise ValueError(f"File {file_name} not found")

    def get_root_folder(self, path_name: str) -> dict:
        """Get the root folder."""
        folders = self.__execute_get_query(f"mimeType='{MimeTypes.FOLDER}'", ["id", "name"])
        for folder in folders:
            if folder["name"] == path_name:
                return folder

    def get_child_folder_by_name(self, root_folder: dict, folder_name: str) -> dict:
        """Get the child folder by name."""
        folders = self.__execute_get_query(
            f"mimeType='{MimeTypes.FOLDER}'  and '{root_folder['id']}' in parents", ["id", "name"]
        )
        for folder in folders:
            if folder["name"] == folder_name:
                return folder
        else:
            raise ValueError(f"Folder {folder_name} not found")

    def create_folder(self, folder_name: str, parent: list[dict]) -> dict:
        """Create the folder."""
        return self.__execute_create_query(file_name=folder_name, mime_type=MimeTypes.FOLDER, folders=parent)

    def create_file(self, file_name, folder, mime_type) -> dict:
        """Create the file."""
        return self.__execute_create_query(file_name=file_name, mime_type=mime_type, folders=[folder])

    @retry(tries=3, delay=1, exceptions=HttpError)
    def __execute_create_query(self, file_name: str, mime_type: str, folders: list[dict]) -> dict:
        try:
            file_metadata = {"name": file_name, "mimeType": mime_type}
            if folders:
                file_metadata["parents"] = [folder["id"] for folder in folders]
            return self.service.files().create(body=file_metadata).execute()
        except HttpError as error:
            logging.error(f"An error occurred during query execution: {error}")
            raise error

    @retry(tries=3, delay=1, exceptions=HttpError)
    def __execute_get_query(self, query: str, fields: list) -> list[dict]:
        files = []
        page_token = None
        while True:
            try:
                response = (
                    self.service.files()
                    .list(
                        q=query,
                        spaces="drive",
                        fields=f"nextPageToken, files({', '.join(fields)})",
                        pageToken=page_token,
                    )
                    .execute()
                )
                files.extend(response.get("files", []))
                page_token = response.get("nextPageToken")
                if not page_token:
                    break
            except HttpError as error:
                logging.error(f"An error occurred during query execution: {error}")
                raise error
        return files

    def lock_file(self, file: dict):
        """Lock the file."""
        self.service.files().update(fileId=file["id"], body={"appProperties": {"locked": "true"}}).execute()

    def unlock_file(self, file: dict):
        """Unlock the file."""
        self.service.files().update(fileId=file["id"], body={"appProperties": {"locked": "false"}}).execute()

    def check_file_locked(self, file: dict) -> bool:
        """Check if the file is locked."""
        file = self.service.files().get(fileId=file["id"], fields="appProperties").execute()
        locked = file.get("appProperties", {}).get("locked", "false")
        return locked == "true"

    @retry(tries=3, exceptions=ValueError)
    def get_file_content(self, file: dict) -> dict:
        """Get the file content."""
        for _ in range(12):
            if self.check_file_locked(file):
                logger.info("File is locked. Waiting for 5 seconds")
                sleep(5)
            else:
                file_content = self.service.files().get_media(fileId=file["id"]).execute()
                self.lock_file(file)
                if file_content:
                    return json.loads(file_content.decode("utf-8"))
                else:
                    return {}
        else:
            self.unlock_file(file)
            raise ValueError("File is locked")



"""Assets module."""
from datetime import datetime, timedelta
from typing import Optional

from coverage import Coverage

from t_qa.config import Inputs
from t_qa.goole_api.google_drive_service import GoogleDriveService
from t_qa.models import TestCase
from t_qa.workitems import RUN_NUMBER


class RunAssets:
    """Assets class."""

    def __init__(self, google_drive_service: GoogleDriveService):
        """Initialize the Assets object."""
        self._google_drive_service = google_drive_service
        self._file = {}
        self._assets = {}
        self._run_data = {}

    def get_assets(self):
        """Get the assets from the Google Drive."""
        self._file = self._google_drive_service.create_file_if_not_exists(
            folder=self._google_drive_service.get_root_folder("T-QA Assets"),
            file_name=Inputs.ADMIN_CODE,
            mime_type="application/json",
        )
        self._assets = self._google_drive_service.get_file_content(file=self._file)
        self._delete_old_run_data()
        self._run_data = self._assets.get(RUN_NUMBER, {})
        return self._assets

    def upload_assets(self) -> None:
        """Upload the assets to the Google Drive."""
        self._google_drive_service.unlock_file(self._file)
        self._assets[RUN_NUMBER] = self._run_data
        self._google_drive_service.update_json_file(file=self._file, data=self._assets)

    def update_test_cases(self, test_cases: list[TestCase]) -> list[TestCase]:
        """Update the test cases."""
        if self._run_data.get("test_cases", False):
            for index, test_case in enumerate(test_cases):
                if test_case.status != "":
                    self._run_data["test_cases"][index]["status"] = test_case.status
        else:
            self._run_data["test_cases"] = [test_case.__dict__ for test_case in test_cases]

        return [TestCase(**test_case) for test_case in self._run_data["test_cases"]]

    def update_bug_counter(self, bug_counter: int) -> int:
        """Update the bug counter."""
        if self._run_data.get("bug_counter", False):
            self._run_data["bug_counter"] += bug_counter
        else:
            self._run_data["bug_counter"] = bug_counter
        return self._run_data["bug_counter"]

    def _delete_old_run_data(self) -> None:
        """Delete old assets."""
        for key, value in self._assets.items():
            if value.get("run_date", None):
                run_date = datetime.strptime(value["run_date"], "%Y-%m-%d %H:%M:%S")
                if run_date < datetime.now() - timedelta(days=3):
                    del self._assets[key]

    def update_record_status_counters(self, record_status_counters: dict[str, int]) -> dict[str, int]:
        """Update the record status counters."""
        if self._run_data.get("record_status_counters", False):
            for key, value in record_status_counters.items():
                if key in self._run_data["record_status_counters"]:
                    self._run_data["record_status_counters"][key] += value
                else:
                    self._run_data["record_status_counters"][key] = value
        else:
            self._run_data["record_status_counters"] = record_status_counters
        return self._run_data["record_status_counters"]

    def update_start_datetime(self, start_datetime: datetime) -> datetime:
        """Update the start datetime."""
        if self._run_data.get("start_datetime", False):
            asset_datetime = datetime.strptime(self._run_data["start_datetime"], "%Y-%m-%d %H:%M:%S")
            if start_datetime < asset_datetime:
                self._run_data["start_datetime"] = start_datetime.strftime("%Y-%m-%d %H:%M:%S")
        else:
            self._run_data["start_datetime"] = start_datetime.strftime("%Y-%m-%d %H:%M:%S")
        return datetime.strptime(self._run_data["start_datetime"], "%Y-%m-%d %H:%M:%S")

    def update_code_coverage_percentage(self, coverage: Coverage) -> str:
        """Update the code coverage."""
        if self._run_data.get("code_coverage", None):
            total_lines = self._run_data["code_coverage"].get("total_lines", 0)
            covered_lines = self._run_data["code_coverage"].get("covered_lines", 0)
        else:
            self._run_data["code_coverage"] = {}
            total_lines = 0
            covered_lines = 0

        for file in coverage.get_data().measured_files():
            analysis = coverage._analyze(file)
            total_lines += len(analysis.statements)
            covered_lines += len(analysis.executed)

        self._run_data["code_coverage"]["total_lines"] = total_lines
        self._run_data["code_coverage"]["covered_lines"] = covered_lines
        percentage = (covered_lines / total_lines) * 100 if total_lines > 0 else 0
        return f"{percentage :.0f}%"

    def get_row_number(self) -> Optional[int]:
        """Get the row number."""
        if self._run_data.get("row_number", False):
            return self._run_data["row_number"]
        else:
            return None

    def set_row_number(self, row_number) -> None:
        """Set the row number."""
        self._run_data["row_number"] = row_number



"""Test file for the QA process."""
import os
from datetime import datetime
from test.secrets import get_attachment

from t_qa.config import SCOPES, Inputs
from t_qa.excel_processing.google_sheet import GoogleSheet
from t_qa.excel_processing.report import _Report
from t_qa.goole_api.account import Account
from t_qa.goole_api.google_drive_service import GoogleDriveService
from t_qa.models import RunData


class TestExcel:
    """Test excel."""

    def setup_method(self):
        """Setup tests."""
        Inputs.ADMIN_CODE = "TST1"
        TEMP_PATH = os.path.join(os.getcwd(), "temp")
        get_attachment(
            "AMH T-QA Google",
            "service_account_key.json",
            TEMP_PATH,
        )
        service_account_path = os.path.join(TEMP_PATH, "service_account_key.json")
        self.run_data = RunData(
            run_date=datetime.now(),
            duration="short",
            empower_env="environment",
            run_link="processRunUrl",
            status="status",
            test_cases=[],
            bugs="Not implemented",
            code_coverage="Not implemented",
            record_status_counters={},
        )
        BOT_ACCOUNT = Account(
            service_account_key_path=service_account_path,
            scopes=SCOPES,
        )
        self.google_sheet = GoogleSheet(BOT_ACCOUNT)
        self.google_drive = GoogleDriveService(BOT_ACCOUNT)

    def test_get_file_id(self):
        """Testing google drive."""
        self.report = _Report("qa.xlsx", google_sheet=self.google_sheet, google_drive=self.google_drive)
        assert self.report.google_sheet_file_id


if __name__ == "__main__":
    test = TestExcel()
    test.setup_method()
    test.test_get_file_id()



"""Columns block for the QA report."""
from .config import DEFAULT_WORK_ITEM_FIELD
from .models import Color, TestCase
from .workitems import VARIABLES


class ColumnsBlock:
    """Columns block for the QA report."""

    def __init__(self, name: str = "", column_names: list[str] = None, color: str = ""):
        """Initialize the columns block."""
        self.name = name
        self.column_names = column_names
        self.color = color


class InputsBlock(ColumnsBlock):
    """Inputs block for the QA report."""

    def __init__(self):
        """Initialize the inputs block."""
        super().__init__(
            name="Inputs",
            column_names=[key for key in VARIABLES.keys() if key not in DEFAULT_WORK_ITEM_FIELD],
            color=Color.inputs_block.value,
        )


class RecordsBlock(ColumnsBlock):
    """Records block for the QA report."""

    def __init__(self, records: dict[str, int]):
        """Initialize the records block."""
        super().__init__(
            name="Records",
            column_names=["Total"],
            color=Color.records_block.value,
        )
        self.column_names.extend(records.keys())


class RunDetailsBlock(ColumnsBlock):
    """Run details block for the QA report."""

    def __init__(self):
        """Initialize the run details block."""
        super().__init__(
            name="Run Details",
            column_names=[
                "Date",
                "Time CST",
                "Duration",
                "Emp Env",
                "Run link",
                "Status",
                "Bugs",
                "Code Cov",
            ],
            color=Color.run_details_block.value,
        )


class TestCasesBlock(ColumnsBlock):
    """Test cases block for the QA report."""

    def __init__(self, test_cases: list[TestCase]):
        """Initialize the test cases block."""
        super().__init__(
            name="Test Cases",
            column_names=[t_c.id for t_c in test_cases],
            color=Color.test_cases_block.value,
        )



"""Models for QA module."""
from datetime import datetime
from enum import Enum


class TestCase:
    """Test case model."""

    def __init__(self, id: str, name: str, status: str = ""):
        """Initialize the test case."""
        self.id = id
        self.status = status
        self.name = name


class RunData:
    """Run data model."""

    def __init__(
        self,
        run_date: datetime,
        duration: str,
        empower_env: str,
        run_link: str,
        status: str,
        bugs: str,
        code_coverage: str,
        test_cases: list[TestCase],
        record_status_counters: dict[str, int],
    ):
        """Initialize the run data."""
        self.run_date: datetime = run_date
        self.duration = duration
        self.empower_env = empower_env
        self.run_link: str = run_link
        self.status = status
        self.bugs = bugs
        self.code_coverage = code_coverage
        self.test_cases: list[TestCase] = test_cases
        self.record_status_counters = record_status_counters


class Color(str, Enum):
    """Colors."""

    white = "FFFFFFFF"
    green = "FFd9ead3"
    red = "FFf3cccb"
    gray = "FFf3f3f3"
    run_details_block = "00b7d7a8"
    records_block = "0099ccff"
    inputs_block = "00ffcc99"
    test_cases_block = "00ffda66"


class Alignment(str, Enum):
    """Alignment."""

    left = "LEFT"
    center = "CENTER"


class DumpCell:
    """Dump cell for the Excel file."""

    def __init__(
        self,
        value: str,
        row: int,
        column_number: int,
        color: str = None,
        alignment: str = Alignment.left.value,
        block_name: str = None,
    ):
        """Initialize the dump cell."""
        self.row = row
        self.block_name = block_name
        self.column_number = column_number
        self.value = value
        self.bg_color = color
        self.alignment = alignment



"""Utility functions for the QA library."""
import sys

from .status import Status


class SingletonMeta(type):
    """The Singleton metaclass."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Call the Singleton class."""
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


def install_sys_hook(t_qa):
    """Installs a system hook to handle unhandled exceptions."""
    if getattr(sys.excepthook, "qa_process", None):
        return

    __sys_excepthook = sys.excepthook

    def excepthook(*exc_info):
        __excepthook(t_qa)

        if __sys_excepthook:
            __sys_excepthook(*exc_info)

    sys.excepthook = excepthook
    sys.excepthook.qa_process = __excepthook


def __excepthook(t_qa) -> None:
    """Handles unhandled exceptions."""
    t_qa.is_run_failed = Status.FAIL.value



"""Module for excel class."""
from openpyxl.workbook import Workbook


class ExcelReport:
    """Excel class represent local excel file."""

    def __init__(self, local_file_path: str) -> None:
        """Init for excel class."""
        self.local_file_path = local_file_path
        self.workbook = Workbook()
        self.workbook.remove(self.workbook.active)
        self.workbook.create_sheet("Report")
        self.worksheet = self.workbook["Report"]

    def write_cells_to_excel(self, all_cells: list) -> None:
        """Write the cells to the Excel file."""
        for cell in all_cells:
            self.worksheet.cell(row=cell.row, column=cell.column_number, value=cell.value)
        self.workbook.save(self.local_file_path)

    def write_header_to_excel(self, headers: list) -> None:
        """Write the header to the Excel file."""
        for cell in headers:
            self.worksheet.cell(row=cell.row, column=cell.column_number, value=cell.value)



"""This module contains the enum that is intended to be used to set the status of anything."""


from enum import Enum


class Status(str, Enum):
    """Status."""

    PASS = "pass"
    FAIL = "fail"
    SUCCESS = "success"



"""WorkItems module to get the work items from Robocorp WorkItems library or from local file."""
try:
    from RPA.Robocorp.WorkItems import WorkItems

    work_items = WorkItems()
    work_items.get_input_work_item()
    work_item = work_items.get_work_item_variables()

    METADATA = work_item.get("metadata", dict())
    RUN_NUMBER = METADATA.get("processRunUrl", "").split("/")[-1]
    VARIABLES = work_item.get("variables", dict())
except (ImportError, KeyError):
    VARIABLES = {}
    METADATA = {}
    RUN_NUMBER = METADATA.get("processRunUrl", "").split("/")[-1]



"""Module for Google services."""
from google.oauth2 import service_account

from ..goole_api.account import Account


class GoogleServices:
    """Google services class."""

    @staticmethod
    def _get_credentials(account: Account) -> service_account.Credentials:
        credentials = service_account.Credentials.from_service_account_file(
            account.service_account_key_path, scopes=account.scopes
        )
        return credentials



"""Top-level package for T QA."""

__author__ = """Thoughtful"""
__email__ = "support@thoughtful.ai"
__version__ = "__version__ = '0.1.14'"

from .qa import configure_qa, test_case_passed, test_case_failed, set_test_records_status_count

__all__ = [
    "set_test_records_status_count",
    "configure_qa",
    "test_case_passed",
    "test_case_failed",
]



"""Module for working with Report."""
import re
import traceback
from typing import Optional

from pytz import timezone

from ..columns_blocks import ColumnsBlock, InputsBlock, RecordsBlock, RunDetailsBlock, TestCasesBlock
from ..config import Inputs
from ..excel_processing.google_sheet import GoogleSheet
from ..exception import AdminCodeException
from ..goole_api.google_drive_service import GoogleDriveService
from ..logger import logger
from ..models import Alignment, Color, DumpCell, RunData, TestCase
from ..status import Status
from ..workitems import VARIABLES
from .excel_report import ExcelReport


class _Report:
    """QA report."""

    def __init__(
        self,
        local_excel_file_path: str,
        google_sheet: GoogleSheet = None,
        google_drive: GoogleDriveService = None,
        last_row: int = None,
    ) -> None:
        """Initialize the Report."""
        self.excel = ExcelReport(local_file_path=local_excel_file_path)
        self.google_sheet = google_sheet
        self.google_drive = google_drive

    def dump(self, run_data: RunData, row_number: Optional[int]) -> Optional[int]:
        """Dump the report."""
        self.google_sheet.set_file_id(self._create_or_get_google_sheet_file())
        headers = self._get_headers(run_data=run_data)
        try:
            self.google_sheet._check_or_create_headers(headers=headers)
        except Exception as e:
            logger.error(f"Error: {e}")
            traceback.print_exc()
            self.excel.write_header_to_excel(headers=headers)

        all_cells = self._get_run_data_dump_cells(run_data=run_data)
        try:
            if not row_number:
                row_number = self.google_sheet.get_last_row()
            self.google_sheet.write_to_google_sheet(all_cells=all_cells, headers=headers, row_number=row_number)
            return row_number
        except Exception as e:
            logger.error(f"Error: {e}")
            traceback.print_exc()
            self.excel.write_cells_to_excel(all_cells=all_cells)

    def _create_or_get_google_sheet_file(self) -> str:
        """Get the file ID."""
        folder, file = self._get_folder_name_and_googlesheet_name()
        google_drive_folder = self.google_drive.create_folder_if_not_exists(folder)
        return self.google_drive.create_file_if_not_exists(google_drive_folder, file)["id"]

    def _get_folder_name_and_googlesheet_name(self) -> tuple[str, str]:
        """Get the folder name and Google Sheet name."""
        try:
            project_admin_code: str = Inputs.ADMIN_CODE
            project_number = re.search(r"([0-9]+)", project_admin_code).group(0)
            folder = project_admin_code.replace(project_number, "").strip().upper()
            file = project_admin_code.strip().upper()
            return folder, file
        except AttributeError:
            raise AdminCodeException("Could not find admin code in work items")

    def get_block_header(self, block: ColumnsBlock, block_start_column=0) -> list[DumpCell]:
        """Write the block header."""
        header_block = []
        for index, column_name in enumerate(block.column_names):
            header_block.append(
                DumpCell(
                    value=column_name,
                    row=1,
                    column_number=block_start_column + index,
                    color=block.color,
                    alignment=Alignment.center.value,
                    block_name=block.name,
                )
            )
        return header_block

    def _get_headers(self, run_data: RunData) -> list[DumpCell]:
        """Write the header."""
        test_cases_block = TestCasesBlock(run_data.test_cases)
        run_details_block = RunDetailsBlock()
        records_block = RecordsBlock(run_data.record_status_counters)
        inputs_block = InputsBlock()

        # Update headers with remote values
        remote_header_block = self.google_sheet.get_remote_header_block()
        if remote_header_block:
            remote_headers: list = remote_header_block[0]
            sub_headers: list = remote_header_block[1]

            run_detail_headers = sub_headers[remote_headers.index("Run Details") : remote_headers.index("Records")]
            records_headers = sub_headers[remote_headers.index("Records") : remote_headers.index("Inputs")]
            inputs_headers = sub_headers[remote_headers.index("Inputs") : remote_headers.index("Test Cases")]
            test_cases_headers = sub_headers[remote_headers.index("Test Cases") :]

            run_details_block.column_names += list(set(run_detail_headers).difference(run_details_block.column_names))
            records_block.column_names += list(set(records_headers).difference(set(records_block.column_names)))
            inputs_block.column_names += list(set(inputs_headers).difference(set(inputs_block.column_names)))
            test_cases_block.column_names += list(set(test_cases_headers).difference(test_cases_block.column_names))

        run_details_start_column = 1
        records_start_column = run_details_start_column + len(run_details_block.column_names)
        inputs_start_column = records_start_column + len(records_block.column_names)
        test_case_start_column = inputs_start_column + len(inputs_block.column_names)

        headers = []
        headers += self.get_block_header(run_details_block, run_details_start_column)
        headers += self.get_block_header(records_block, records_start_column)
        headers += self.get_block_header(inputs_block, inputs_start_column)
        headers += self.get_block_header(test_cases_block, test_case_start_column)
        if not remote_header_block:
            self.google_sheet.get_header_block(run_details_block, run_details_start_column)
            self.google_sheet.get_header_block(records_block, records_start_column)
            self.google_sheet.get_header_block(inputs_block, inputs_start_column)
            self.google_sheet.get_header_block(test_cases_block, test_case_start_column)
        return headers

    def _get_run_data_dump_cells(self, run_data: RunData) -> list[DumpCell]:
        """Dump the run data."""
        row = 2
        headers = self.google_sheet.get_remote_header_block()[1]
        all_cells = []
        all_cells += self.get_run_details_cells(run_data, row, headers)
        all_cells += self.get_records_cells(run_data, row, headers)
        all_cells += self.get_inputs_cells(row, headers)
        all_cells += self.get_test_cases_cells(run_data.test_cases, row, headers)
        return all_cells

    def get_run_details_cells(self, run_data: RunData, row: int, headers: list[str]):
        """Get the run details cells."""
        run_date = run_data.run_date.astimezone(timezone("US/Central"))
        return [
            DumpCell(
                value=run_data.run_link,
                row=row,
                column_number=self.get_column_number("Run link", headers),
                block_name="Run Details",
            ),
            DumpCell(
                value=run_date.strftime("%Y-%m-%d"),
                row=row,
                column_number=self.get_column_number("Date", headers),
                block_name="Run Details",
            ),
            DumpCell(
                value=run_date.strftime("%H:%M:%S"),
                row=row,
                column_number=self.get_column_number("Time CST", headers),
                block_name="Run Details",
            ),
            DumpCell(
                value=run_data.status,
                row=row,
                column_number=self.get_column_number("Status", headers),
                block_name="Run Details",
            ),
            DumpCell(
                value=run_data.bugs,
                row=row,
                column_number=self.get_column_number("Bugs", headers),
                block_name="Run Details",
            ),
            DumpCell(
                value=run_data.code_coverage,
                row=row,
                column_number=self.get_column_number("Code Cov", headers),
                block_name="Run Details",
            ),
            DumpCell(
                value=run_data.empower_env,
                row=row,
                column_number=self.get_column_number("Emp Env", headers),
                block_name="Run Details",
            ),
            DumpCell(
                value=run_data.duration,
                row=row,
                column_number=self.get_column_number("Duration", headers),
                block_name="Run Details",
            ),
        ]

    def get_records_cells(self, run_data: RunData, row: int, headers: list[DumpCell]):
        """Get the records cells."""
        total_records = sum([counter for counter in run_data.record_status_counters.values()]) or 0
        record_block_headers = RecordsBlock(run_data.record_status_counters).column_names
        dump_cells = [
            DumpCell(
                value=str(total_records),
                row=row,
                column_number=self.get_column_number("Total", headers),
                block_name="Records",
            ),
        ]
        # add record if it exists in remote file and not exists in record_status_counters
        for header in record_block_headers:
            if header != "Total" and header not in run_data.record_status_counters.keys():
                dump_cells.append(
                    DumpCell(
                        value="",
                        row=row,
                        column_number=self.get_column_number(header, headers),
                        block_name="Records",
                    )
                )

        for status, counter in run_data.record_status_counters.items():
            dump_cells.append(
                DumpCell(
                    value=str(counter),
                    row=row,
                    column_number=self.get_column_number(status, headers),
                    block_name="Records",
                ),
            )
        return dump_cells

    def get_inputs_cells(self, row: int, headers: list[DumpCell]):
        """Get the inputs cells."""
        input_block = InputsBlock().column_names
        cells = []
        for inputs in input_block:
            cells.append(
                DumpCell(
                    value=VARIABLES.get(inputs, "Empty"),
                    row=row,
                    column_number=self.get_column_number(inputs, headers),
                    block_name="Inputs",
                )
            )
        return cells

    def get_test_cases_cells(self, test_cases: list[TestCase], row: int, headers: list[DumpCell]):
        """Get the test cases cells."""
        status_colors = {Status.FAIL.value: Color.red.value, Status.PASS.value: Color.green.value, "": Color.gray.value}
        cells = []
        for test_case in test_cases:
            cells.append(
                DumpCell(
                    value=test_case.status,
                    row=row,
                    column_number=self.get_column_number(test_case.id, headers),
                    color=status_colors[test_case.status],
                    alignment="CENTER",
                    block_name="Test Cases",
                )
            )
        return cells

    def get_column_number(self, header_name: str, headers: list[str]) -> int:
        """Get the column number."""
        for index, cell in enumerate(headers, start=1):
            if cell and cell == header_name:
                return index
        else:
            raise ValueError(f"Header {header_name} not found")



"""Configuration file for the QA library."""
import os

from t_qa.workitems import METADATA, VARIABLES

DEFAULT_WORK_ITEM_FIELD = [
    "userEmail",
    "changeStatusUrl",
    "environment",
    "supervisorCallbackUrl",
    "jwtProcessUpdatesWebhook",
    "refreshTokenUrl",
    "processRunUrl",
    "accessToken",
    "refreshToken",
]
ROOT_PATH = os.getcwd()
OUTPUT_PATH = os.path.join(ROOT_PATH, "output")
TEMP_PATH = os.path.join(ROOT_PATH, "temp")
if os.path.exists(OUTPUT_PATH):
    DEFAULT_QA_RESULT_FILE_PATH = os.path.join(OUTPUT_PATH, "qa_result.xlsx")
else:
    DEFAULT_QA_RESULT_FILE_PATH = os.path.join(ROOT_PATH, "qa_result.xlsx")
DEFAULT_TEST_CASES_FILE_PATH = os.path.join(ROOT_PATH, "test_cases.yaml")
LOCAL_RUN = os.environ.get("RC_PROCESS_RUN_ID") is None
SCOPES = ["https://www.googleapis.com/auth/drive", "https://www.googleapis.com/auth/spreadsheets"]


class Inputs:
    """Inputs class."""

    ADMIN_CODE: str = METADATA.get("process", dict()).get("adminCode", "")
    ENVIRONMENT: str = VARIABLES.get("environment", "")



"""Exception Module fot T-QA process."""


class TQaBaseException(Exception):
    """Base exception for qa lib."""

    pass


class TQaBaseSilentException(TQaBaseException):
    """Base Exception with no logging messages."""

    pass


class SkipIfItLocalRunException(TQaBaseSilentException):
    """Skip process if it is local run."""

    pass


class SkipIfItWorkItemsNotSetException(TQaBaseException):
    """Skip process if it is local run."""

    pass


class NotLoggedInToGoogleException(TQaBaseSilentException):
    """Exception raise when lib could not log in to google services."""

    pass


class SkipProdRunException(TQaBaseException):
    """Exception to skip the prod run."""

    pass


class TestCaseFileDoesNotExistException(TQaBaseException):
    """Exception to raise when test cases file does not exist."""

    pass


class ServiceAccountKeyPathException(TQaBaseException):
    """Exception raise if lib could not get service account key."""

    pass


class TestCaseReadException(TQaBaseException):
    """Exception raise if lib could not read Yaml file."""

    pass


class AdminCodeException(TQaBaseException):
    """Exception raise if lib could not get Admin code from work items."""

    pass



"""QA.py module contains the QA class for the Digital Worker's process."""
import atexit
import os
import re
import traceback
from datetime import datetime
from typing import Optional

import pytz
import t_bug_catcher
import yaml
from coverage import Coverage
from ta_bitwarden_cli.exceptions import NotLoggedInError
from ta_bitwarden_cli.handler import get_attachment
from thoughtful.supervisor import report_builder

from .config import DEFAULT_QA_RESULT_FILE_PATH, DEFAULT_TEST_CASES_FILE_PATH, LOCAL_RUN, SCOPES
from .excel_processing.google_sheet import GoogleSheet
from .excel_processing.report import _Report
from .exception import (
    NotLoggedInToGoogleException,
    ServiceAccountKeyPathException,
    SkipIfItLocalRunException,
    SkipIfItWorkItemsNotSetException,
    TestCaseFileDoesNotExistException,
    TestCaseReadException,
    TQaBaseException,
    TQaBaseSilentException,
)
from .goole_api.account import Account
from .goole_api.google_drive_service import GoogleDriveService
from .logger import logger
from .models import RunData, TestCase
from .run_assets import RunAssets
from .status import Status
from .utils import SingletonMeta, install_sys_hook
from .workitems import METADATA, RUN_NUMBER, VARIABLES


class QA(metaclass=SingletonMeta):
    """QA class for the Digital Worker's process."""

    def __init__(self) -> None:
        """Initialize the QA process."""
        if not LOCAL_RUN:
            if Coverage.current():
                self.coverage = Coverage.current()
            elif Coverage().config.config_file:
                self.coverage = Coverage()
            else:
                self.coverage = Coverage(branch=True)
                self.coverage.exclude(r"    def __repr__")
                self.coverage.exclude(r"raise AssertionError")
                self.coverage.exclude(r"raise NotImplementedError")
                self.coverage.exclude(r"if 0:")
                self.coverage.exclude(r"if __name__ == .__main__.:")
                self.coverage.exclude(r"if TYPE_CHECKING:")
                self.coverage.exclude(r"class .*\bProtocol\):")
                self.coverage.exclude(r"@(abc\.)?abstractmethod")
            self.coverage.start()
        self.run_status: str = Status.SUCCESS.value
        self.service_account_key_path: Optional[str] = None
        self.test_cases: list[TestCase] = []
        self.record_status_counters = {}

    def configurate(
        self,
        test_cases_file_path: str = DEFAULT_TEST_CASES_FILE_PATH,
        service_account_key_path: str = None,
    ) -> None:
        """Configurate the QA process."""
        try:
            self._set_test_cases(test_cases_file_path)
            self._set_service_account_key_path(service_account_key_path)
            self._skip_if_it_local_run()
        except TQaBaseSilentException:
            return
        except TQaBaseException as e:
            logger.warning(e)
            return
        self._set_start_datetime()
        atexit.register(self.dump)

    def test_case_pass(self, id: str) -> None:
        """Check the test case passed."""
        self._set_test_case_status(id=id, status=Status.PASS.value)

    def test_case_fail(self, id: str) -> None:
        """Check the test case failed."""
        self._set_test_case_status(id=id, status=Status.FAIL.value)

    def dump(self):
        """Dump the test cases."""
        try:
            self._get_code_coverage()
            account = Account(
                service_account_key_path=self.service_account_key_path,
                scopes=SCOPES,
            )
            google_drive_service = GoogleDriveService(account)
            run_assets = RunAssets(google_drive_service)
            run_assets.get_assets()

            self.test_cases = run_assets.update_test_cases(self.test_cases)
            self.start_datetime = run_assets.update_start_datetime(self.start_datetime)
            self.record_status_counters = run_assets.update_record_status_counters(self.record_status_counters)

            bugs = run_assets.update_bug_counter(t_bug_catcher.get_errors_count())
            code_coverage = run_assets.update_code_coverage_percentage(self.coverage)

            run_data = RunData(
                run_date=self.start_datetime,
                duration=self.__get_duration(),
                empower_env="Prod" if VARIABLES.get("environment", "") == "production" else "Dev",
                run_link=f'=HYPERLINK("{METADATA.get("processRunUrl", "")}", "{RUN_NUMBER}")',
                status=self._get_run_result(),
                test_cases=self.test_cases,
                bugs=str(bugs),
                code_coverage=code_coverage,
                record_status_counters=self.record_status_counters,
            )
            report = _Report(
                local_excel_file_path=DEFAULT_QA_RESULT_FILE_PATH,
                google_sheet=GoogleSheet(account),
                google_drive=google_drive_service,
            )
            row_number = report.dump(
                run_data=run_data,
                row_number=run_assets.get_row_number(),
            )
            run_assets.set_row_number(row_number)
            run_assets.upload_assets()
        except TQaBaseException as e:
            logger.warning(e)
        except Exception as e:
            logger.error(f"Error during dumping: {e}")
            traceback.print_exc()

    def set_test_records_status_count(self, record_status_counters: dict[str, int]):
        """Set the test records status count."""
        if len(record_status_counters.keys()) >= 15:
            logger.error("The record status count is too large. It should be less than 15 statuses.")
            record_status_counters = {}
        self.record_status_counters = record_status_counters

    def _get_run_result(self):
        if self.run_status == Status.SUCCESS.value:
            try:
                self.run_status = report_builder.status.value
            except AttributeError:
                logger.warning("Could not get the run result from supervisor.")
        return self.run_status

    def __get_duration(self):
        duration = datetime.now(pytz.UTC) - self.start_datetime.astimezone(pytz.UTC)
        seconds = duration.seconds
        minutes = seconds // 60
        hours = minutes // 60
        duration_str = f"{hours}h {minutes % 60}m {seconds % 60}s"
        return duration_str

    def _set_test_cases(self, test_cases_file_path: str) -> None:
        if not os.path.exists(test_cases_file_path):
            raise TestCaseFileDoesNotExistException(f"Test cases file not found: {test_cases_file_path}")
        try:
            with open(test_cases_file_path) as test_cases_file:
                test_cases = yaml.safe_load(test_cases_file)["test_cases"]
                self.test_cases = [TestCase(**test_case) for test_case in test_cases]
        except (TypeError, KeyError, ValueError) as e:
            raise TestCaseReadException(f"Error during reading test cases: {e}")

    def _set_service_account_key_path(self, service_account_key_path: Optional[str]) -> None:
        if service_account_key_path:
            self.service_account_key_path = service_account_key_path
        else:
            try:
                self.service_account_key_path = get_attachment(
                    "T-QA Google",
                    "service_account_key.json",
                )
            except NotLoggedInError:
                raise NotLoggedInToGoogleException()
            except ValueError:
                raise ServiceAccountKeyPathException("There are no access to 'T-QA Google' collection")

    def _skip_if_it_local_run(self):
        if LOCAL_RUN:
            raise SkipIfItLocalRunException()
        if not METADATA:
            raise SkipIfItWorkItemsNotSetException("'Metadata' is not set")
        if not VARIABLES:
            raise SkipIfItWorkItemsNotSetException("'Variables' is not set")

    def _set_start_datetime(self) -> None:
        """Get the start datetime."""
        try:
            root_path = os.environ.get("ROBOT_ROOT", "")
            console_log_folder_path = os.path.abspath(os.path.join(root_path, os.pardir))
            console_log_file_path = os.path.join(console_log_folder_path, "console.txt")
            with open(console_log_file_path, "r") as file:
                data = file.read()
            date_str = re.findall(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}", data)[0]
            date_str += "UTC"
            self.start_datetime = datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%S%Z")
        except (TypeError, FileNotFoundError, IndexError) as e:
            logger.warning("Could not get the start datetime from the empower.")
            self.start_datetime = datetime.now()
            t_bug_catcher.report_error(exception=e)

    def _set_test_case_status(self, id: str, status: str):
        """Check the test case."""
        for test_case in self.test_cases:
            if test_case.id == id:
                test_case.status = status

    def _get_code_coverage(self):
        self.coverage.stop()
        self.coverage.save()
        self.coverage.report(
            ignore_errors=True,
            omit=["*/t_qa/*"],
        )


t_qa = QA()
install_sys_hook(t_qa)

configure_qa = t_qa.configurate
test_case_failed = t_qa.test_case_fail
test_case_passed = t_qa.test_case_pass
set_test_records_status_count = t_qa.set_test_records_status_count
